from math import sqrt, ceil
from numpy import int32, float32

from .linalgcore import LinearAlgebraCore
from .linalgcore import dot_kernel



_dot = dot_kernel()

def dot( rtrn, ary1, ary2, stream=None ):
    shp1 = ary1.shape
    shp2 = ary2.shape
    shpr = rtrn.shape

    if ( ( shp1[1] != shp2[0] ) or ( shpr != (shp1[0],shp2[1]) ) ):
        raise ValueError( "size doesn't fit" )

    tx = ary1.tx
    it = ary1.it
    
    _block = (tx, 1, 1)
    _grid  = (shp1[0],shp2[1])

    _dot(
        rtrn.gpudata, it,
        ary1.gpudata, ary1.n_row, ary1.n_col,
        ary2.gpudata, ary2.n_row, ary2.n_col,
        block=_block, grid=_grid
    )


class LinearAlgebra:
    core = LinearAlgebraCore()

    def __init__(self):
        pass

    def dot(self, result, ary1, ary2, stream=None):
        shp1 = ary1.shape
        shp2 = ary2.shape
        shpr = result.shape
        if ((shp1[1] != shp2[0]) or (shpr != (shp1[0],shp2[1]))):
            raise ValueError("size doesn't fit")
        ## optimal kernel size
        tx = int(sqrt(shp1[1]))
        _block = (tx, 1, 1)
        _grid  = (shp1[0],shp2[1])

        iters = int32(ceil(shp1[1] / tx))

        self.__class__.core.dot(
            result, iters,
            ary1, ary1.n_row, ary1.n_col,
            ary2, ary2.n_row, ary2.n_col,
            block=_block, grid=_grid, stream=stream
        )

    def norm(self, ary, axis=None, result=None, stream=None):
        shp = ary.shape

        self.__class__.core.norm(
        )

    """
    def vstack(self, *arys, result=None):
        if (result == cudg.array):
            shpr = result.shape
        rows = 0
        for ary in arys:
            shpi = ary.shape
            if (shpr[1] != shpi[1]):
                raise ValueError("NO")
            rows += shpi[0]
        if (shpr[0] != rows):
            raise ValueError("NO")

        self.__class__.core.vstack(
            result
        )
    """



from VRAManager import VRAM, cuda

from numpy import ndarray
from numpy import array, empty
from numpy import int32, float32



class gpusparse:


    def __init__( self, shape:tuple, gpudata: ndarray ):
        '''
        GPU Sparse Array Type
        '''
        self.shape = shape

        self.n_row = int32( shape[0] )
        self.n_col = int32( shape[1] )

        _sparse = self._be_sparse( gpudata )

        self.ary_size = _sparse.shape[1]

        self.idxdata = VRAM._request( self.ary_size )
        self.gpudata = VRAM._request( self.ary_size )


    def __del__( self ):
        pass


    def _be_sparse( self, gpudata:ndarray ):
        n_row = self.shape[0]
        n_col = self.shape[1]
        tol = 10e-38
        idx = []
        val = []

        for row in n_row:
            for col in n_col:
                val_rc = gpudata[row,col]
                if ( abs( val_rc < tol ) ):
                    idx.append( row * n_col + col )
                    val.append( val_rc )
        
        return array( [idx, val], dtype=float32 )

    
    def _to_gpu( self, gpudata: ndarray ):

        pass

    
    def to_gpu( self, gpudata: ndarray ):

        _sparse = self._be_sparse( gpudata )

        if ( self.ary_size == _sparse.shape[1] ):

            cuda.memcpy_htod( self.idxdata, _sparse[0,:].astype(int32)   )
            cuda.memcpy_htod( self.gpudata, _sparse[1,:].astype(float32) )


import pycuda.driver as cuda



class Memory:

    type = cuda.DeviceAllocation


    def _request( self, size: int ) -> cuda.DeviceAllocation:
        '''
        for each number has 4 byte 
        because GPU uses 32bit
        '''
        return cuda.mem_alloc( size * 4 )

    
    def _filter( self, gpudata ):
        pass


    def _return( self, mem: cuda.DeviceAllocation ) -> None:

        try:
            mem.free()
        except:
            print( 'Memory Link might be happened' )
            print( 'Recommand to restart kernel' )


VRAM = Memory()


from pycuda.compiler import SourceModule

def dot_kernel():
    ## block=(tx,1,1), grid=(row,col)
    kernel_code = \
    """
    #define tx (threadIdx.x)
    #define bx (blockIdx.x)
    #define by (blockIdx.y)
    #define N_ (blockDim.x)

    __global__ void dot(
        float* rtrn, int iters,
        float* ary1, int n_row1, int n_col1,
        float* ary2, int n_row2, int n_col2
    ) {
        __shared__ float sM[1024];
        sM[tx] = 0;

        int idxi;
        int idx1;
        int idx2;
        int idxr;

        for ( int i = 0; i < iters; i++ ) {
            idxi = tx * iters + i;

            if ( idxi < n_row2 ) {
                idx1 = bx * n_col1 + idxi;
                idx2 = by + n_col2 * idxi;

                sM[tx] += ( ary1[idx1] * ary2[idx2] );
            }
        }
        __syncthreads();

        if ( tx == 0 ) {
            idxr = bx * n_col2 + by;
            rtrn[idxr] = 0;
            for ( int i = 0; i < N_; i++ ) {
                rtrn[idxr] += sM[i];
            }
        }
        __syncthreads();
    }
    """
    kernel = SourceModule( kernel_code )
    
    return kernel.get_function( "dot" )


class LinearAlgebraCore:

    def __init__(self):
        self.dot = self.dot_kernel()
        # self.norm = self.norm_kernel()

    def dot_kernel(self):
        ## block=(tx,1,1), grid=(row,col)
        kernel_code = \
        """
        #define tx (threadIdx.x)
        #define bx (blockIdx.x)
        #define by (blockIdx.y)
        #define N_ (blockDim.x)

        __global__ void dot(
            float* result, int iteration,
            float* ary1, int n_row1, int n_col1,
            float* ary2, int n_row2, int n_col2
        ) {
            __shared__ float sM[1024];
            sM[tx] = 0;

            int idxi;
            int idx1;
            int idx2;
            int idxr;

            for (int i = 0; i < iteration; i++) {
                idxi = tx * iteration + i;

                if (idxi < n_row2) {
                    idx1 = bx * n_col1 + idxi;
                    idx2 = by + n_col2 * idxi;

                    sM[tx] += (ary1[idx1] * ary2[idx2]);
                }
            }
            __syncthreads();

            if (tx == 0) {
                idxr = bx * n_col2 + by;
                result[idxr] = 0;
                for (int i = 0; i < N_; i++) {
                    result[idxr] += sM[i];
                }
            }
            __syncthreads();
        }
        """
        kernel = SourceModule(kernel_code)
        
        return kernel.get_function("dot")

    def norm_kernel(self):
        ## block=(n_row,1,1), grid=(1,1,1)
        kernel_code = \
        """
        __device__ float _sqrt(float value) {
            float s = 0;
            float t = 0;

            s = value / 2;

            for (;s != t;) {
                t = s;
                s = ((value/t) + t) / 2;
            }

            return s;
        }

        __global__ void _norm(
            float* result, 
        ) {
        }
        """
        kernel = SourceModule(kernel_code)

        return kernel.get_function("_norm")



from ..VRAManager import VRAM, cuda

from numpy import ndarray
from numpy import empty
from numpy import int32, float32

from math import sqrt, ceil

from functools import reduce



class gpuarray:


    def __init__( self, shape: tuple, gpudata: ndarray ):
        '''
        GPU Array Type
        '''
        self.shape    = shape
        self.ary_size = reduce( lambda x, y: x * y, shape )

        self.n_row = int32( shape[0] )
        self.n_col = int32( shape[1] )

        self.gpudata = VRAM._request( self.ary_size )

        self.to_gpu( gpudata )

        self.kernel_size()

    
    def __del__( self ):

        VRAM._return( self.gpudata )

    
    def to_gpu( self, gpudata: ndarray ):

        if isinstance( gpudata, ndarray ):
            if ( gpudata.itemsize == 4 ):
                pass
            elif ( gpudata.itemsize == 8 ):
                gpudata = gpudata.astype( float32 )
            else:
                raise ValueError( 'somethings strange' )
        else:
            raise ValueError( 'not numpy.ndarray' )

        cuda.memcpy_htod( self.gpudata, gpudata )

    
    def to_host( self ):
        gpudata = empty( ( self.ary_size ), dtype=float32 )

        cuda.memcpy_dtoh( gpudata, self.gpudata )

        return gpudata.reshape( self.shape )


    def kernel_size( self ):

        bx = int( self.n_row / 32 ) + 1
        by = int( self.n_col / 32 ) + 1

        self._block = (32,32,1)
        self._grid  = (bx,by,1)

        self.tx = int( sqrt( self.shape[1] ) )
        self.it = int32( ceil( self.shape[1] / self.tx ) )